<?php
// $Id: pifr_simpletest.client.inc,v 1.36 2010/01/19 02:06:39 boombatower Exp $

/**
 * @file
 * Provide SimpleTest client review implementation.
 *
 * @author Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

module_load_include('client.inc', 'pifr_drupal');

/**
 * SimpleTest PIFR review implementation.
 */
class pifr_client_review_pifr_simpletest extends pifr_client_review_pifr_drupal {

  /**
   * Add clone-db URL to list of files to be downloaded (if set).
   */
  public function __construct(array $test) {
    parent::__construct($test);

    if (!empty($this->test['review']['argument']['clone-db'])) {
      $this->test['files'][] = $this->test['review']['argument']['clone-db'];
    }
  }

  /**
   * Import clone database (if set) before performing Drupal installation.
   */
  protected function install() {
    parent::install();

    if (!empty($this->test['review']['argument']['clone-db'])) {
      // Drupal install() will have taken care of settings.php and such so drop
      // and create database to be filled by clone-db.
      $this->database->drop_database();
      $this->database->create_database();

      // Since a database import is being used the username and password must
      // be included in arguments.
      if (!empty($this->test['review']['argument']['clone-db-user']) &&
          !empty($this->test['review']['argument']['clone-db-pass'])) {
        $this->admin_username = $this->test['review']['argument']['clone-db-user'];
        $this->admin_password = $this->test['review']['argument']['clone-db-pass'];
      }
      else {
        $this->set_error(array('@reason' => 'both [clone-db-user] and [clone-db-pass] must be specified'));
        return;
      }

      $basename = basename($this->test['review']['argument']['clone-db']);
      foreach ($this->files as $file) {
        if (basename($file) == $basename) {
          $this->log('Importing clone database from [' . $basename . ']...');

          if ($result = $this->database->import($file, 'clone_db')) {
            $this->log('Imported [' . $result . '] tables.');
          }
          else {
            $this->set_error(array('@reason' => 'failed to import clone database'));
          }
          break;
        }
      }
    }
  }

  /**
   * Run SimpleTest tests.
   */
  protected function review() {
    global $base_url, $base_path;

    // If Drupal 6 then we need to perform SimpleTest installation.
    if ($this->test['review']['argument']['core'] == 6) {
      if (!$this->test_prepare()) {
        return;
      }
    }

    if (!$this->database->set_variable('simpletest_clear_results', FALSE)) {
      $this->set_error(array('@reason' => 'failed to set SimpleTest results variable in checkout database'));
    }
    if (!$this->database->set_variable('simpletest_verbose', FALSE)) {
      $this->set_error(array('@reason' => 'failed to set SimpleTest verbose variable in checkout database'));
    }

    // Login to Drupal HEAD and enable the SimpleTest module.
    $base_url .= '/checkout';
    $base_path = '/checkout';

    $b = new PIFRBrowser();

    // Login as admin.
    $user = new stdClass();
    $user->name = $this->admin_username;
    $user->pass_raw = $this->admin_password;
    $b->drupalLogin($user);

    // Enable SimpleTest.
    $edit = array();
    if ($this->test['review']['argument']['core'] == 6) {
      // Drupal 6.
      $edit['status[simpletest]'] = TRUE;
      $b->drupalPost('admin/build/modules', $edit, t('Save configuration'));
    }
    else {
      // Drupal 7 and above.
      $edit['modules[Core][simpletest][enable]'] = TRUE;
      $b->drupalPost('admin/modules', $edit, t('Save configuration'));
    }

    $base_url = $this->original['base_url'];
    $base_path = $this->original['base_path'];

    if (($b->results['#exception'] + $b->results['#fail']) != 0) {
      $this->set_error(array('@reason' => 'failed to enable SimpleTest'));
      return;
    }

    // Run tests from command line.
    chdir($this->checkout_directory);

    $url = escapeshellarg(url('', array('absolute' => TRUE)) . 'checkout');
    if (!$this->exec(PIFR_CLIENT_PHP . ' ./scripts/run-tests.sh --concurrency ' . PIFR_CLIENT_CONCURRENCY .
      ' --php ' . PIFR_CLIENT_PHP . ' --url ' . $url . ' ' . $this->test_list())) {
      $this->set_error(array('@reason' => t('failed during invocation of run-tests.sh')));
      return;
    }

    // Get test info for use by test_info_parse().
    if (!$this->exec(PIFR_CLIENT_PHP . ' ./scripts/run-tests.sh --php ' . PIFR_CLIENT_PHP . ' --url ' . $url . ' --list')) {
      $this->set_error(array('@reason' => t('failed attempting to get list of tests from run-tests.sh')));
      return;
    }

    chdir($this->original['directory']);
  }

  /**
   * Perform Drupal 6 SimpleTest installation.
   */
  protected function test_prepare() {
    // Apply core patch.
    $files = array(
      $this->checkout_directory . '/' . $this->module_directory . '/simpletest/D6-core-simpletest.patch',
    );
    parent::apply(NULL, $files);

    // Move run-tests.sh to scripts directory.
    chdir($this->checkout_directory);
    if (!copy($this->module_directory . '/simpletest/run-tests.sh', 'scripts/run-tests.sh')) {
      $this->set_error(array('@reason' => 'failed to install run-tests.sh'));
      return FALSE;
    }
    chdir($this->original['directory']);

    $this->log('Completed Drupal 6 SimpleTest installation.');

    return TRUE;
  }

  /**
   * Create the test argument for run-tests.sh.
   *
   * @return string Tests argument to be appended to run-test.sh command.
   */
  protected function test_list() {
    if (!empty($this->test['review']['argument']['tests'])) {
      return '--class ' . implode(',', $this->test['review']['argument']['tests']);
    }
    elseif (!empty($this->test['review']['argument']['modules'])) {
      $args = array();
      foreach ($this->test['review']['argument']['modules'] as $module) {
        $file = $this->module_path($module) . '/' . $module . '.test';
        if (file_exists($file)) {
          $args[] = $this->module_path($module) . '/' . $module . '.test';
        }
      }
      return '--file ' . implode(',', $args);
    }
    elseif (!empty($this->test['review']['argument']['directory'])) {
      $scan = file_scan_directory($this->test['review']['argument']['directory'], '\.test$');

      $files = array();
      foreach ($scan as $file) {
        $files[] = $file->filename;
      }
      return '--file ' . implode(',', $files);
    }
    elseif (PIFR_DEBUG) {
      return '--class NonDefaultBlockAdmin';
    }
    return '--all';
  }

  /**
   * Add the SimpleTest specific results.
   */
  public function get_result() {
    $result = parent::get_result();

    // If result code is one greater then number of operations then all the
    // operations passed, so the test results need to be added.
    $fail_code = count($this->operations) + 1;
    if ($result['code'] == $fail_code) {
      $result['details'] = array(
        '@pass' => 0,
        '@fail' => 0,
        '@exception' => 0,
      );
      $result['data'] = array();

      // Cycle through assertions to fill in results.
      $info = $this->test_info_parse();

      // Process assertions in batches of 10,000 to keep down memory usage.
      $message_id = 0;
      do {
        $assertions = $this->database->query('SELECT *
                                              FROM {simpletest}
                                              WHERE message_id > ' . $message_id . '
                                              ORDER BY message_id
                                              LIMIT 10000');
        $non_pass = 0;
        foreach ($assertions as $assertion) {
          // Fill in details with summary of assertion counts.
          $result['details']['@' . $assertion['status']]++;

          // Initialize each test class.
          if (!isset($result['data'][$assertion['test_class']])) {
            $result['data'][$assertion['test_class']] = array(
              'test_name' => !empty($info[$assertion['test_class']]) ? $info[$assertion['test_class']] : $assertion['test_class'],
              'pass' => 0,
              'fail' => 0,
              'exception' => 0,
              'assertions' => array(),
            );
          }

          // Add up summary totals for test class.
          $result['data'][$assertion['test_class']][$assertion['status']]++;

          // Store non-pass assertion until reaching maximum number.
          if ($assertion['status'] != 'pass' && $non_pass < PIFR_CLIENT_MAX_ASSERTIONS) {
            // Remove non-standard keys.
            unset($assertion['message_id']);
            unset($assertion['test_id']);

            // Make file only include name, instead of absolute reference.
            $assertion['file'] = basename($assertion['file']);

            // Add assertion to set of non-pass assertions for the class.
            $result['data'][$assertion['test_class']]['assertions'][] = $assertion;

            // Keep a record of the number of non-pass assertions.
            $non_pass++;
          }
        }

        $message_id += 10000;
      }
      while ($assertions);

      // Remove class name keys.
      $result['data'] = array_values($result['data']);

      // Assign the code bassed on the test results.
      $result['code'] = ($result['details']['@fail'] + $result['details']['@exception'] ? $fail_code : $fail_code + 1);
    }
    return $result;
  }

  /**
   * Parse test info from 'run-tests.sh --list' output.
   *
   * @return array Associative array of test info keyed by test case.
   */
  protected function test_info_parse() {
    $this->output = array_slice($this->output, 4);
    $group = 'Unknown';
    $info = array();
    foreach ($this->output as $line) {
      if (isset($line[1]) && $line[1] == '-') {
        if (preg_match('/^ - (.*?) \((\w+)\)$/m', $line, $match)) {
          $info[$match[2]] = $match[1] . ' (' . $match[2] . ') [' . $group . ']';
        }
      }
      else {
        $group = $line;
      }
    }
    return $info;
  }

  /**
   * If on review operation then display the assertion count.
   */
  public static function status() {
    if (variable_get('pifr_client_operation', 'init') == 'review') {
      module_load_include('review.inc', 'pifr_client');
      $type = pifr_client_review_database_type();

      require_once drupal_get_path('module', 'pifr_client') . '/review/db/db.inc';
      require_once drupal_get_path('module', 'pifr_client') . '/review/db/' . $type. '.inc';

      $database = 'pifr_client_db_interface_' . $type;
      $database = new $database;

      $count = $database->query('SELECT COUNT(*) FROM {simpletest}');
      return t('Running tests, @count assertion(s).', array('@count' => number_format($count ? $count[0]['COUNT(*)'] : 0)));
    }
    return parent::status();
  }
}
