<?php
// $Id: pifr_server.xmlrpc.inc,v 1.39 2010/01/18 21:33:36 boombatower Exp $

/**
 * @file
 * Provide XML-RPC callbacks and invoke functions.
 *
 * @author Jimmy Berry ("boombatower", http://drupal.org/user/214218)
 */

/*
 * Maximum number of items to send in a batch.
 */
define('PIFR_SERVER_BATCH_MAX_COUNT', 100);

/**
 * Queue a the batch of files for testing.
 *
 * @param string $client_key Client key requesting batch to be queued.
 * @param array $batch Array of file and project data used to queue files.
 * <code>
 *   $batch = array(
 *     'branches' => array( [referenced branches (may also be tested)] ),
 *     'files' => array( [files to be tested] ),
 *     'projects' => array( [referenced projects] ),
 *   );
 *
 *   $branch = array(
 *     'project_identifier', // Reference to project by client identifier.
 *     'client_identifier', // Identifier that will be used to reference the branch.
 *     'vcs_identifier', // Version control system identifier.
 *     'dependency', // Comma seperated list of branch identifiers that this branch depends on.
 *     'plugin_argument', // Associative array of arguments keyed by plugin.
 *     'test', // Request branch to be tested.
 *   );
 *
 *   $file = array(
 *     'branch_identifier', // Reference to branch by client identifier.
 *     'client_identifier', // Identifier that will be used to reference the file.
 *     'file_url', // Absolute URL to file.
 *   );
 *
 *   $project = array(
 *     'client_identifier', // Identifier that will be used to reference the project.
 *     'name', // Name of project.
 *     'repository_type', // Repository type (currently only 'cvs').
 *     'repository_url', // Repository URL, pserver:anonymous:anonymous@cvs.drupal.org:/cvs/drupal.
 *   );
 * </code>
 * @return array Assigned test IDs, or array with response code, PIFR_RESPONSE_*.
 * <code>
 *   $response = array(
 *     'branches' => array( [Test IDs of branches keyed by client_identifier] ),
 *     'files' => array( [Test IDs of files keyed by client_identifier] ),
 *   );
 * </code>
 *
 * OR
 *
 * <code>
 *   $test = array(
 *     'response' // Response code: PIFR_RESPONSE_*.
 *   );
 * </code>
 */
function pifr_server_xmlrpc_queue($client_key, $batch) {
  pifr_debug('XML-RPC batch(): received.');

  // Check if valid project client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_PROJECT);

  if ($client) {
    pifr_debug('XML-RPC batch(): valid project client.');

    if (!PIFR_ACTIVE) {
      pifr_debug('XML-RPC batch(): server is not active, request denied.');
      return array(
        'response' => PIFR_RESPONSE_DENIED,
      );
    }

    // Store project data and retrieve completed project information.
    $projects = pifr_server_xmlrpc_project_save($batch['projects'], $client['client_id']);

    // Store branch data and retrieve completed branch information.
    $branches = pifr_server_xmlrpc_branch_save($batch['branches'], $projects, $client);

    // Store branch data and retrieve completed file information.
    $files = pifr_server_xmlrpc_file_save($batch['files'], $branches, $client);

    // Postpone tests if branch is currently failing.
    foreach ($branches as $branch) {
      if (!pifr_server_result_summary_status($branch['test_id'])) {
        // Postpone all tests that have been queued.
        pifr_server_file_branch_update($branch['branch_id'], PIFR_SERVER_TEST_STATUS_QUEUED, PIFR_SERVER_TEST_STATUS_POSTPONED);
      }
    }

    // Send back test IDs.
    $response = array(
      'branches' => array(),
      'files' => array(),
    );

    foreach ($branches as $branch) {
      $response['branches'][$branch['client_identifier']] = $branch['test_id'];
    }

    foreach ($files as $file) {
      $response['files'][$file['client_identifier']] = $file['test_id'];
    }

    return $response;
  }
  else {
    pifr_debug('XML-RPC batch(): invalid project client.');
    return array(
      'response' => PIFR_RESPONSE_INVALID_SERVER,
    );
  }
}

/**
 * Save specified projects.
 *
 * @param array $projects Projects to save.
 * @param integer $client_id Client ID that send projects.
 * @return array Saved projects.
 */
function pifr_server_xmlrpc_project_save(array $projects, $client_id) {
  $saved = array();
  foreach ($projects as $project) {
    // Add client ID to complete project information.
    $project['client_id'] = $client_id;

    // Save the project and add complete project information to saved array.
    $saved[$project['client_identifier']] = pifr_server_project_save($project);
  }
  return $saved;
}

/**
 * Save specified branches.
 *
 * @param array $branches Branches to save.
 * @param array $projects Projects that branches relate to.
 * @param array $client Client information.
 * @return array Saved branches.
 */
function pifr_server_xmlrpc_branch_save(array $branches, array $projects, array $client) {
  $saved = array();
  foreach ($branches as $branch) {
    if (!isset($projects[$branch['project_identifier']])) {
      pifr_debug(t('XML-RPC batch(): missing project data for branch: (b: @identifier; p: @project_identifier).',
                   array('@identifier' => $branch['client_identifier'], '@project_identifier' => $branch['client_identifier'])));
      continue;
    }

    // Get project ID from saved project information.
    $branch['project_id'] = $projects[$branch['project_identifier']]['project_id'];

    // Save the branch and add complete branch information to saved array.
    $saved[$branch['client_identifier']] = pifr_server_branch_save($branch);

    // If branch is to be tested then update test record.
    if ($branch['test']) {
      $test = pifr_server_test_get($saved[$branch['client_identifier']]['test_id']);
      pifr_server_xmlrpc_test_queue($test, $client);
    }
  }

  // Once all branches have been saved the branch dependencies can be safely
  // converted to branch ID references.
  foreach ($saved as $branch) {
    if (!empty($branch['dependency'])) {
      $dependencies = explode(',', $branch['dependency']);
      $dependency_ids = array();
      foreach ($dependencies as $dependency) {
        if (!empty($saved[$dependency])) {
          $dependency_ids[] = $saved[$dependency]['branch_id'];
        }
      }

      $branch['dependency'] = implode(',', $dependency_ids);
      pifr_server_branch_save($branch);
    }
  }

  return $saved;
}

/**
 * Save specified files.
 *
 * @param array $files Files to save.
 * @param array $branches Branches that files relate to.
 * @param array $client Client information.
 * @return array Saved files.
 */
function pifr_server_xmlrpc_file_save(array $files, array $branches, array $client) {
  $saved = array();
  foreach ($files as $file) {
    if (!isset($branches[$file['branch_identifier']])) {
      pifr_debug(t('XML-RPC batch(): missing branch data for file: (u: @file_url; n: @issue_nid).',
                   array('@file_url' => $file['file_url'], '@issue_nid' => $file['issue_nid'])));
      continue;
    }

    // Get branch ID from saved branch information.
    $file['branch_id'] = $branches[$file['branch_identifier']]['branch_id'];

    // Generate file name from file URL.
    $file['file_name'] = basename($file['file_url']);

    // Save the file and add complete file information to saved array.
    $saved[$file['client_identifier']] = pifr_server_file_save($file);

    // Update test record.
    $test = pifr_server_test_get($saved[$file['client_identifier']]['test_id']);
    pifr_server_xmlrpc_test_queue($test, $client);
  }
  return $saved;
}

/**
 * Queue test for testing.
 *
 * If test is already queued or currently being tested the queue request will
 * be ignored and an appropriate log message will be made.
 *
 * @param array $test Test information.
 * @param array $client Client information.
 */
function pifr_server_xmlrpc_test_queue(array $test, array $client) {
  if ($test['status'] == PIFR_SERVER_TEST_STATUS_QUEUED) {
    pifr_server_log($test['test_id'], $client['client_id'], PIFR_SERVER_LOG_TEST_REQUEST_IGNORE);
  }
  else {
    pifr_server_test_queue($test);

    pifr_server_log($test['test_id'], $client['client_id'], PIFR_SERVER_LOG_TEST_REQUEST);
  }
}

/**
 * Request the next file for testing. (Called by PIFR client)
 *
 * @param string $client_key Client key requesting next test.
 * @return array Test information, empty array, or array with response code, PIFR_RESPONSE_*.
 * <code>
 *   $test = array(
 *     'test_id', // Unique test ID.
 *     'review' => array(
 *       'plugin', // Module name of the review plugin to use.
 *       'argument', // Associative array of arguments for the review backend.
 *     ),
 *     'vcs' => array(
 *       'main' => array(
 *         'repository' => array(
 *            'type', // Textual representation of repository type.
 *            'url', // CVS will only require URL.
 *         ),
 *         'vcs_identifier', // Version control system identifier.
 *       ),
 *     ),
 *     'dependencies' => array(), // Dependencies following the same format as 'main'.
 *     'files' => array(), // List of absolute URL(s) to file(s) necessary for the review.
 *   );
 * </code>
 *
 * OR
 *
 * <code>
 *   $test = array(
 *     'response' // Response code: PIFR_RESPONSE_*.
 *   );
 * </code>
 */
function pifr_server_xmlrpc_next($client_key) {
  pifr_debug('XML-RPC next(): received.');

  // Check if valid test client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_TEST);

  if ($client) {
    pifr_debug('XML-RPC next(): valid test client.');

    if (!PIFR_ACTIVE) {
      pifr_debug('XML-RPC next(): server is not active, request denied.');
      return array(
        'response' => PIFR_RESPONSE_DENIED,
      );
    }

    return pifr_server_test_get_next($client);
  }
  else {
    pifr_debug('XML-RPC next(): invalid test client.');
    return array(
      'response' => PIFR_RESPONSE_INVALID_SERVER,
    );
  }
  return array(
    'response' => PIFR_RESPONSE_ACCEPTED,
  );
}

/**
 * Report test results. (Called by PIFR client)
 *
 * @param string $client_key Client key sending test results.
 * @param array $result Test result.
 * <code>
 *   $result = array(
 *     'test_id', // Unique test ID.
 *     'code', // Summary result code as defined by the plugin steps.
 *     'details', // Array of details (parameters for t()).
 *     'pass', // Total number of pass assertions.
 *     'fail', // Total number of fail assertions.
 *     'exception', // Total number of exception assertions.
 *     'result_details' => array( [summary of test and non-pass assertions] ),
 *   );
 *
 *   $result_detail = array(
 *     'test_name', // Name of test.
 *     'pass', // Total number of pass assertions.
 *     'fail', // Total number of fail assertions.
 *     'exception', // Total number of exception assertions.
 *     'assertions' => array( [non-pass assertions] ),
 *   );
 *
 *   $assertion = array(
 *     'status', // Assertion status: 'fail' or 'exception'.
 *     'message', // Descriptive assertion message.
 *     'message_group', // Message group.
 *     'function', // Function in which the assertion was triggered.
 *     'line', // Line at which the assertion was triggered.
 *     'file', // File in which the assertion was triggered.
 *   );
 * </code>
 * @return Response code: PIFR_RESPONSE_*.
 */
function pifr_server_xmlrpc_result($client_key, array $result) {
  pifr_debug('XML-RPC result(): received.');

  // Check if valid test client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_TEST);

  if ($client) {
    pifr_debug('XML-RPC result(): valid test client.');

    // Ensure that test was sent to the client returning the result.
    $test = pifr_server_test_get($result['test_id']);
    if ($test['status'] == PIFR_SERVER_TEST_STATUS_QUEUED && !pifr_server_environment_status_get_client($test, $client)) {
      watchdog('pifr_server', 'Test not assigned to client: (t: @test_id, c: @client_id)',
               array('@test_id' => $test['test_id'], '@client_id' => $client['client_id']), WATCHDOG_ALERT);
      return PIFR_RESPONSE_DENIED;
    }

    pifr_server_test_result($test, $result, $client);
  }
  else {
    pifr_debug('XML-RPC result(): invalid test client.');
    return PIFR_RESPONSE_INVALID_SERVER;
  }
  return PIFR_RESPONSE_ACCEPTED;
}

/**
 * Retrieve test results. (Called by PIFT)
 *
 * @param string $client_key Client key requesting batch to be queued.
 * @param integer $since GMT Unix timestamp to query test results after.
 * @return array Test results, empty array, or array with response code, PIFR_RESPONSE_*.
 * <code>
 *   $batch = array(
 *     'results' => array( [results of tests] ),
 *     'next', // More results are available, query using the provided GMT Unix timestamp.
 *   );
 *
 *   $result = array(
 *     'test_id', // Unique ID representing the test.
 *     'pass', // TRUE or FALSE.
 *     'message', // Message describing results of test.
 *   );
 * </code>
 *
 * OR
 *
 * <code>
 *   $test = array(
 *     'response' // Response code: PIFR_RESPONSE_*.
 *   );
 * </code>
 */
function pifr_server_xmlrpc_retrieve($client_key, $since) {
  pifr_debug('XML-RPC retrieve(): received.');

  // Check if valid test client.
  $client = pifr_server_client_check_key($client_key, PIFR_SERVER_CLIENT_TYPE_PROJECT);

  if ($client) {
    pifr_debug('XML-RPC retrieve(): valid project client.');

    if (!PIFR_SERVER_REPORT) {
      pifr_debug('XML-RPC retrieve(): reporting is not enabled, request denied.');
      return array(
        'response' => PIFR_RESPONSE_DENIED,
      );
    }

    $batch = array(
      'results' => array(),
      'next' => FALSE,
    );

    $count = 0;
    $since = pifr_server_xmlrpc_time_local($since);
    $tests = pifr_server_test_get_since($since);
    foreach ($tests as $test) {
      if ($count >= PIFR_SERVER_BATCH_MAX_COUNT) {
        $batch['next'] = pifr_server_xmlrpc_time_gmt($test['last_tested']);
        break;
      }

      $batch['results'][] = array(
        'test_id' => $test['test_id'],
        'pass' => pifr_server_result_summary_status($test['test_id']),
        'message' => pifr_server_result_summary_message($test['test_id']),
      );
      pifr_server_log($test['test_id'], $client['client_id'], PIFR_SERVER_LOG_CLIENT_RETRIEVE);
      $count++;
    }

    return $batch;
  }
  else {
    pifr_debug('XML-RPC retrieve(): invalid test client.');
    return array(
      'response' => PIFR_RESPONSE_INVALID_SERVER,
    );
  }
  return array(
    'response' => PIFR_RESPONSE_ACCEPTED,
  );
}

/**
 * Convert a GMT Unix timestamp to a local Unix timestamp.
 *
 * @param integer $gmt_timestamp GMT Unix timestamp.
 * @return integer Local timestamp.
 */
function pifr_server_xmlrpc_time_local($gmt_timestamp) {
  $gmt_now = gmmktime();
  $offset = $gmt_now - $gmt_timestamp;
  return time() - $offset;
}

/**
 * Convert a local Unix timestamp to a GMT Unix timestamp.
 *
 * @param integer $timestamp Local Unix timestamp.
 * @return integer GMT Unix timestamp
 */
function pifr_server_xmlrpc_time_gmt($timestamp) {
  $now = time();
  $offset = $now - $timestamp;
  return gmmktime() - $offset;
}
